// Copyright 2023 Memgraph Ltd.
//
// Use of this software is governed by the Business Source License
// included in the file licenses/BSL.txt; by using this file, you agree to be bound by the terms of the Business Source
// License, and you may not use this file except in compliance with the Business Source License.
//
// As of the Change Date specified in that file, in accordance with
// the Business Source License, use of this software will be governed
// by the Apache License, Version 2.0, included in the file
// licenses/APL.txt.

#pragma once

#include <cstdint>
#include <memory>

#include "storage/v3/edge_ref.hpp"
#include "storage/v3/id_types.hpp"
#include "storage/v3/property_value.hpp"
#include "storage/v3/vertex.hpp"
#include "storage/v3/vertex_id.hpp"
#include "utils/logging.hpp"
#include "utils/synchronized.hpp"

namespace memgraph::storage::v3 {

// Forward declarations because we only store pointers here.
struct Edge;
struct Delta;
struct CommitInfo;

inline uint64_t GetNextDeltaUUID() noexcept {
  static utils::Synchronized<uint64_t, utils::SpinLock> delta_id{0};
  return delta_id.WithLock([](auto id) { return id++; });
}

// This class stores one of three pointers (`Delta`, `Vertex` and `Edge`)
// without using additional memory for storing the type. The type is stored in
// the pointer itself in the lower bits. All of those structures contain large
// items in themselves (e.g. `uint64_t`) that require the pointer to be aligned
// to their size (for `uint64_t` it is 8). That means that the pointer will
// always be a multiple of 8 which implies that the lower 3 bits of the pointer
// will always be 0. We can use those 3 bits to store information about the type
// of the pointer stored (2 bits).
class PreviousPtr {
 private:
  static constexpr uintptr_t kDelta = 0b01UL;
  static constexpr uintptr_t kVertex = 0b10UL;
  static constexpr uintptr_t kEdge = 0b11UL;

  static constexpr uintptr_t kMask = 0b11UL;

 public:
  enum class Type {
    NULLPTR,
    DELTA,
    VERTEX,
    EDGE,
  };

  struct Pointer {
    Pointer() = default;
    explicit Pointer(Delta *delta) : type(Type::DELTA), delta(delta) {}
    explicit Pointer(Vertex *vertex) : type(Type::VERTEX), vertex(vertex) {}
    explicit Pointer(Edge *edge) : type(Type::EDGE), edge(edge) {}

    Type type{Type::NULLPTR};
    Delta *delta{nullptr};
    Vertex *vertex{nullptr};
    Edge *edge{nullptr};
  };

  PreviousPtr() {}

  PreviousPtr(const PreviousPtr &other) noexcept : storage_(other.storage_) {}
  PreviousPtr(PreviousPtr &&) = delete;
  PreviousPtr &operator=(const PreviousPtr &) = delete;
  PreviousPtr &operator=(PreviousPtr &&) = delete;
  ~PreviousPtr() = default;

  Pointer Get() const {
    if (storage_ == 0) {
      return {};
    }
    uintptr_t type = storage_ & kMask;
    if (type == kDelta) {
      // NOLINTNEXTLINE(performance-no-int-to-ptr)
      return Pointer{reinterpret_cast<Delta *>(storage_ & ~kMask)};
    }
    if (type == kVertex) {
      // NOLINTNEXTLINE(performance-no-int-to-ptr)
      return Pointer{reinterpret_cast<Vertex *>(storage_ & ~kMask)};
    }
    if (type == kEdge) {
      // NOLINTNEXTLINE(performance-no-int-to-ptr)
      return Pointer{reinterpret_cast<Edge *>(storage_ & ~kMask)};
    }
    LOG_FATAL("Invalid pointer type!");
  }

  void Set(Delta *delta) {
    auto value = reinterpret_cast<uintptr_t>(delta);
    MG_ASSERT((value & kMask) == 0, "Invalid pointer!");
    storage_ = value | kDelta;
  }

  void Set(Vertex *vertex) {
    auto value = reinterpret_cast<uintptr_t>(vertex);
    MG_ASSERT((value & kMask) == 0, "Invalid pointer!");
    storage_ = value | kVertex;
  }

  void Set(Edge *edge) {
    auto value = reinterpret_cast<uintptr_t>(edge);
    MG_ASSERT((value & kMask) == 0, "Invalid pointer!");
    storage_ = value | kEdge;
  }

 private:
  uintptr_t storage_{0};
};

inline bool operator==(const PreviousPtr::Pointer &a, const PreviousPtr::Pointer &b) {
  if (a.type != b.type) return false;
  switch (a.type) {
    case PreviousPtr::Type::VERTEX:
      return a.vertex == b.vertex;
    case PreviousPtr::Type::EDGE:
      return a.edge == b.edge;
    case PreviousPtr::Type::DELTA:
      return a.delta == b.delta;
    case PreviousPtr::Type::NULLPTR:
      return b.type == PreviousPtr::Type::NULLPTR;
  }
}

inline bool operator!=(const PreviousPtr::Pointer &a, const PreviousPtr::Pointer &b) { return !(a == b); }

struct Delta {
  enum class Action : uint8_t {
    // Used for both Vertex and Edge
    DELETE_OBJECT,
    RECREATE_OBJECT,
    SET_PROPERTY,

    // Used only for Vertex
    ADD_LABEL,
    REMOVE_LABEL,
    ADD_IN_EDGE,
    ADD_OUT_EDGE,
    REMOVE_IN_EDGE,
    REMOVE_OUT_EDGE,
  };

  // Used for both Vertex and Edge
  struct DeleteObjectTag {};
  struct RecreateObjectTag {};
  struct SetPropertyTag {};

  // Used only for Vertex
  struct AddLabelTag {};
  struct RemoveLabelTag {};
  struct AddInEdgeTag {};
  struct AddOutEdgeTag {};
  struct RemoveInEdgeTag {};
  struct RemoveOutEdgeTag {};

  Delta(DeleteObjectTag /*unused*/, CommitInfo *commit_info, uint64_t delta_id, uint64_t command_id)
      : action(Action::DELETE_OBJECT), uuid(delta_id), commit_info(commit_info), command_id(command_id) {}

  Delta(RecreateObjectTag /*unused*/, CommitInfo *commit_info, uint64_t delta_id, uint64_t command_id)
      : action(Action::RECREATE_OBJECT), uuid(delta_id), commit_info(commit_info), command_id(command_id) {}

  Delta(AddLabelTag /*unused*/, LabelId label, CommitInfo *commit_info, uint64_t delta_id, uint64_t command_id)
      : action(Action::ADD_LABEL), uuid(delta_id), commit_info(commit_info), command_id(command_id), label(label) {}

  Delta(RemoveLabelTag /*unused*/, LabelId label, CommitInfo *commit_info, uint64_t delta_id, uint64_t command_id)
      : action(Action::REMOVE_LABEL), uuid(delta_id), commit_info(commit_info), command_id(command_id), label(label) {}

  Delta(SetPropertyTag /*unused*/, PropertyId key, const PropertyValue &value, CommitInfo *commit_info,
        uint64_t delta_id, uint64_t command_id)
      : action(Action::SET_PROPERTY),
        uuid(delta_id),
        commit_info(commit_info),
        command_id(command_id),
        property({key, value}) {}

  Delta(AddInEdgeTag /*unused*/, EdgeTypeId edge_type, VertexId vertex_id, EdgeRef edge, CommitInfo *commit_info,
        uint64_t delta_id, uint64_t command_id)
      : action(Action::ADD_IN_EDGE),
        uuid(delta_id),
        commit_info(commit_info),
        command_id(command_id),
        vertex_edge({edge_type, std::move(vertex_id), edge}) {}

  Delta(AddOutEdgeTag /*unused*/, EdgeTypeId edge_type, VertexId vertex_id, EdgeRef edge, CommitInfo *commit_info,
        uint64_t delta_id, uint64_t command_id)
      : action(Action::ADD_OUT_EDGE),
        uuid(delta_id),
        commit_info(commit_info),
        command_id(command_id),
        vertex_edge({edge_type, std::move(vertex_id), edge}) {}

  Delta(RemoveInEdgeTag /*unused*/, EdgeTypeId edge_type, VertexId vertex_id, EdgeRef edge, CommitInfo *commit_info,
        uint64_t delta_id, uint64_t command_id)
      : action(Action::REMOVE_IN_EDGE),
        uuid(delta_id),
        commit_info(commit_info),
        command_id(command_id),
        vertex_edge({edge_type, std::move(vertex_id), edge}) {}

  Delta(RemoveOutEdgeTag /*unused*/, EdgeTypeId edge_type, VertexId vertex_id, EdgeRef edge, CommitInfo *commit_info,
        uint64_t delta_id, uint64_t command_id)
      : action(Action::REMOVE_OUT_EDGE),
        uuid(delta_id),
        commit_info(commit_info),
        command_id(command_id),
        vertex_edge({edge_type, std::move(vertex_id), edge}) {}

  Delta(const Delta &) = delete;
  Delta(Delta &&) = delete;
  Delta &operator=(const Delta &) = delete;
  Delta &operator=(Delta &&) = delete;

  ~Delta() {
    switch (action) {
      case Action::DELETE_OBJECT:
      case Action::RECREATE_OBJECT:
      case Action::ADD_LABEL:
      case Action::REMOVE_LABEL:
        break;
      case Action::ADD_IN_EDGE:
      case Action::ADD_OUT_EDGE:
      case Action::REMOVE_IN_EDGE:
      case Action::REMOVE_OUT_EDGE:
        std::destroy_at(&vertex_edge.vertex_id);
        break;
      case Action::SET_PROPERTY:
        property.value.~PropertyValue();
        break;
    }
  }

  Action action;
  uint64_t uuid;
  // TODO: optimize with in-place copy
  CommitInfo *commit_info;
  uint64_t command_id;
  PreviousPtr prev;
  Delta *next{nullptr};

  union {
    LabelId label;
    struct {
      PropertyId key;
      PropertyValue value;
    } property;
    struct {
      EdgeTypeId edge_type;
      VertexId vertex_id;
      EdgeRef edge;
    } vertex_edge;
  };
};

static_assert(alignof(Delta) >= 8, "The Delta should be aligned to at least 8!");

}  // namespace memgraph::storage::v3
